/*
 * Copyright 2008 Gerd Ziegler (www.gerdziegler.de) Licensed under the Apache
 * License, Version 2.0 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.apache.org/licenses/LICENSE-2.0 Unless required by applicable law
 * or agreed to in writing, software distributed under the License is
 * distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the specific language
 * governing permissions and limitations under the License. @author
 * www.gerdziegler.de
 */
package org.ztest.analyze;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import org.ztest.classinfo.ZIClassEdge;
import org.ztest.classinfo.ZIClassInfo;
import org.ztest.classinfo.ZIClassSetEdgeMeta;
import org.ztest.classinfo.ZIPath;
import org.ztest.classinfo.impl.ZDependencyEdgeMeta;
import org.ztest.graph.ZIEdge;
import org.ztest.graph.ZIGraph;
import org.ztest.graph.jgrapht.ZJGraphTGraph;
import org.ztest.jdom.ZWriteXML;
import org.ztest.sets.ZIClassSet;

public class ZAnalyzeClassSetDependency {
    public static ZIGraph<ZIClassSet, ZIClassSetEdgeMeta, ZWriteXML> computeDependencyGraph(ZIClassInfo classInfo, Set<ZIClassSet> sets)
        throws Exception {
        ZIGraph<ZIClassSet, ZIClassSetEdgeMeta, ZWriteXML> ret = new ZJGraphTGraph<ZIClassSet, ZIClassSetEdgeMeta, ZWriteXML>();

        Map<String, Set<ZIClassSet>> class2ClassSetMap = computeClass2ClassSetMap(sets);
        Map<String, Set<String>> successorCache = new HashMap<String, Set<String>>();

        for (ZIClassEdge e : classInfo.getEdges()) {
            String class1 = e.getSource();
            String class2 = e.getTarget();

            if (class1.equals(class2)) {
                continue;
            }

            Set<ZIClassSet> l1 = class2ClassSetMap.get(class1);
            if (l1 == null) {
                continue;
            }

            Set<String> successors = getLayeredSuccessors(classInfo, class2ClassSetMap, successorCache, class2);

            for (String successor : successors) {
                computeActualDependencyGraph(classInfo, ret, class2ClassSetMap, class1, class2, successor);
            }
        }

        return ret;
    }

    private static Set<String> getLayeredSuccessors(ZIClassInfo classInfo, Map<String, Set<ZIClassSet>> layerMap, Map<String, Set<String>> layeredSuccessorCache, String t1)
        throws Exception {
        Set<String> ret = layeredSuccessorCache.get(t1);
        if (ret != null) {
            return ret;
        }

        if (layerMap.containsKey(t1)) {
            ret = new HashSet<String>();
            ret.add(t1);
            layeredSuccessorCache.put(t1, ret);
            return ret;
        }

        ret = new HashSet<String>();
        layeredSuccessorCache.put(t1, ret);
        Set<String> successors = classInfo.getSuccessors(t1, false);
        for (String crt : successors) {
            if (!crt.equals(t1)) {
                Set<String> crtSet = getLayeredSuccessors(classInfo, layerMap, layeredSuccessorCache, crt);
                ret.addAll(crtSet);
            }
        }

        return ret;
    }

    private static void computeActualDependencyGraph(ZIClassInfo classInfo, ZIGraph<ZIClassSet, ZIClassSetEdgeMeta, ZWriteXML> actualDependencyGraph, Map<String, Set<ZIClassSet>> dependencyMap, String t1, String t2, String successor)
        throws Exception {
        Set<ZIClassSet> l1 = dependencyMap.get(t1);
        Set<ZIClassSet> l2 = dependencyMap.get(successor);

        for (ZIClassSet set1 : l1) {
            for (ZIClassSet set2 : l2) {
                if (!set1.equals(set2)) {
                    ZIEdge<ZIClassSet, ZIClassSetEdgeMeta> actualEdge = actualDependencyGraph.addEdge(set1, set2);
                    // if (!allowedDependencyGraph.isEdge(set1, set2)) {
                    ZIPath path = classInfo.getShortestPath(t2, successor);
                    ZIClassEdge edge = classInfo.getEdge(t1, t2);
                    path.getEdges().add(0, edge);
                    ZIClassSetEdgeMeta meta = actualEdge.getMetaData();
                    if (meta == null) {
                        meta = new ZDependencyEdgeMeta();
                        actualEdge.setMetaData(meta);
                    }
                    meta.getPaths().add(path);
                    // }
                }
            }
        }
    }

    private static Map<String, Set<ZIClassSet>> computeClass2ClassSetMap(Set<ZIClassSet> classSets)
        throws Exception {
        Map<String, Set<ZIClassSet>> ret = new HashMap<String, Set<ZIClassSet>>();
        for (ZIClassSet set : classSets) {
            for (String clazz : set.getClasses()) {
                Set<ZIClassSet> dependencySet = ret.get(clazz);
                if (dependencySet == null) {
                    dependencySet = new HashSet<ZIClassSet>();
                    ret.put(clazz, dependencySet);
                }
                dependencySet.add(set);
            }
        }
        return ret;
    }

}
