/*
 * Copyright 2008 Gerd Ziegler (www.gerdziegler.de) Licensed under the Apache
 * License, Version 2.0 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.apache.org/licenses/LICENSE-2.0 Unless required by applicable law
 * or agreed to in writing, software distributed under the License is
 * distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the specific language
 * governing permissions and limitations under the License. @author
 * www.gerdziegler.de
 */
package org.ztest.classinfo.impl;

import java.io.Externalizable;
import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.jdom.Element;
import org.ztest.classinfo.ZIClassEdge;
import org.ztest.classinfo.ZIClassEdgeFilter;
import org.ztest.classinfo.ZIClassEdgeMeta;
import org.ztest.classinfo.ZIClassFilter;
import org.ztest.classinfo.ZIClassInfo;
import org.ztest.classinfo.ZIClassMeta;
import org.ztest.classinfo.ZIManifest;
import org.ztest.classinfo.ZIPath;
import org.ztest.classinfo.ZPath;
import org.ztest.graph.ZIEdge;
import org.ztest.graph.ZIGraph;
import org.ztest.graph.jgrapht.ZJGraphTGraph;
import org.ztest.jdom.ZXMLUtil;

public class ZClassInfo implements Externalizable, ZIClassInfo {
    private static final long serialVersionUID = 1L;
    // private static final Logger log = Logger.getLogger(ZClassInfo.class);

    private final ZIGraph<String, ZIClassEdgeMeta, ZIClassMeta> graph = new ZJGraphTGraph<String, ZIClassEdgeMeta, ZIClassMeta>();
    private final Map<String, ZLocation> locations = new HashMap<String, ZLocation>();

    public void writeExternal(ObjectOutput out)
        throws IOException {

        out.writeObject(locations);

        List<String> nodes = new ArrayList<String>(getClasses());
        Collections.sort(nodes);
        out.writeInt(nodes.size());
        for (String name : nodes) {
            out.writeObject(name);
            out.writeObject(getClassMeta(name));
        }

        List<ZIClassEdge> edges = new ArrayList<ZIClassEdge>(getEdges());
        Collections.sort(edges);
        out.writeInt(edges.size());
        for (ZIClassEdge edge : edges) {
            out.writeObject(edge.getSource());
            out.writeObject(edge.getTarget());
            out.writeObject(edge.getMetaData());
        }
    }

    public void readExternal(ObjectInput in)
        throws IOException, ClassNotFoundException {

        Map<String, ZLocation> loc = (Map<String, ZLocation>)in.readObject();
        locations.putAll(loc);

        int nodesSize = in.readInt();
        for (int i = 0; i < nodesSize; i++) {
            String name = (String)in.readObject();
            ZIClassMeta meta = (ZIClassMeta)in.readObject();
            graph.addNode(name);
            graph.getNodeMeta(name).put(ZClassMeta.class.getName(), meta);
        }

        int edgesSize = in.readInt();
        for (int i = 0; i < edgesSize; i++) {
            String src = (String)in.readObject();
            String tgt = (String)in.readObject();
            ZIClassEdgeMeta meta = (ZIClassEdgeMeta)in.readObject();
            ZIEdge<String, ZIClassEdgeMeta> edge = graph.addEdge(src, tgt);
            edge.setMetaData(meta);
        }
    }

    private static final ZIClassEdgeFilter FILTER_INSTANCEOF = new ZIClassEdgeFilter() {
        public boolean accept(ZIClassEdge edge) {
            return edge.getMetaData().isInheritance();
        }
    };

    public ZClassInfo() {
    }

    public ZIClassMeta getClassMeta(String name) {
        Map<String, ZIClassMeta> meta = graph.getNodeMeta(name);
        if (meta == null) {
            return null;
        }
        ZIClassMeta ret = (ZIClassMeta)meta.get(ZClassMeta.class.getName());
        if (ret == null) {
            ret = new ZClassMeta(name);
            meta.put(ZClassMeta.class.getName(), ret);
        }
        return ret;
    }

    public boolean isInstanceof(String name, String superClassName)
        throws Exception {
        return name.equals(superClassName) || isSuccessor(name, superClassName, FILTER_INSTANCEOF, true);
    }

    public Set<String> getClassesInstanceof(String className)
        throws Exception {
        Set<String> ret = getPredecessors(className, FILTER_INSTANCEOF, true);
        ret.add(className);
        return ret;
    }

    public Set<String> getSuperclasses(String className)
        throws Exception {
        return getSuccessors(className, FILTER_INSTANCEOF, true);
    }

    public boolean isAnnotatedWith(String className, String annotationClassName)
        throws Exception {
        ZIClassMeta cm = getClassMeta(className);
        if (cm.getAnnotations().contains(annotationClassName)) {
            return true;
        }
        return false;
    }

    public Set<String> getClassesAnnotatedWith(final String annotationClassName)
        throws Exception {
        ZIClassFilter filter = new ZIClassFilter() {
            public boolean accept(String name)
                throws Exception {
                ZIClassMeta cm = getClassMeta(name);
                if (cm.getAnnotations().contains(annotationClassName)) {
                    return true;
                }
                return false;
            }
        };
        Set<String> ret = getClasses(filter);
        return ret;
    }

    public Set<String> getClasses(ZIClassFilter filter)
        throws Exception {
        Set<String> ret = new HashSet<String>();
        for (String s : graph.getNodeSet()) {
            if (filter.accept(s)) {
                ret.add(s);
            }
        }
        return ret;
    }

    public Set<String> getPredecessors(Set<String> classes, boolean transitive)
        throws Exception {
        return graph.getPredecessors(classes, null, transitive);
    }

    public Set<String> getSuccessors(Set<String> classes, boolean transitive)
        throws Exception {
        return graph.getSuccessors(classes, null, transitive);
    }

    public Set<String> getScannedClasses()
        throws Exception {
        Set<String> scanned = getClasses(new ZIClassFilter() {
            public boolean accept(String name)
                throws Exception {
                return getClassMeta(name).isScanned();
            }
        });
        return scanned;
    }

    public ZIClassEdge addEdge(String class1, String class2)
        throws Exception {
        if (!graph.isNode(class1)) {
            throw new Exception("add node first: " + class1);
        }
        if (!graph.isNode(class2)) {
            throw new Exception("add node first: " + class2);
        }
        ZIEdge<String, ZIClassEdgeMeta> e = graph.addEdge(class1, class2);
        ZIClassEdge ret = new ZClassEdge(e);
        return ret;
    }

    public ZIClassEdge addEdgeAndNode(String class1, String class2)
        throws Exception {
        ZIEdge<String, ZIClassEdgeMeta> e = graph.addEdge(class1, class2);
        ZIClassEdge ret = new ZClassEdge(e);
        return ret;
    }

    public void addClass(String name) {
        if (!graph.isNode(name)) {
            graph.addNode(name);
        }
    }

    public List<Set<String>> getCycles() {
        return graph.getCycles();
    }

    public Set<String> getClasses() {
        return graph.getNodeSet();
    }

    public Set<ZIClassEdge> getEdges() {
        Set<ZIClassEdge> ret = new HashSet<ZIClassEdge>();
        for (ZIEdge<String, ZIClassEdgeMeta> e : graph.getEdgeSet()) {
            ret.add(new ZClassEdge(e));
        }
        return ret;
    }

    public Set<ZIClassEdge> getPredecessorEdges(String name, boolean transitive)
        throws Exception {
        return getPredecessorEdges(name, null, transitive);
    }

    public Set<ZIClassEdge> getPredecessorEdges(String name, ZIClassEdgeFilter filter, boolean transitive)
        throws Exception {
        Set<ZIClassEdge> ret = new HashSet<ZIClassEdge>();
        for (ZIEdge<String, ZIClassEdgeMeta> e : graph.getPredecessorEdges(name, ZClassEdgeFilter.create(filter), transitive)) {
            ret.add(new ZClassEdge(e));
        }
        return ret;
    }

    public Set<ZIClassEdge> getSuccessorEdges(String name, ZIClassEdgeFilter filter, boolean transitive)
        throws Exception {
        Set<ZIClassEdge> ret = new HashSet<ZIClassEdge>();
        for (ZIEdge<String, ZIClassEdgeMeta> e : graph.getSuccessorEdges(name, ZClassEdgeFilter.create(filter), transitive)) {
            ret.add(new ZClassEdge(e));
        }
        return ret;
    }

    public Set<ZIClassEdge> getSuccessorEdges(Set<String> set, ZIClassEdgeFilter filter, boolean transitive)
        throws Exception {
        Set<ZIClassEdge> ret = new HashSet<ZIClassEdge>();
        for (ZIEdge<String, ZIClassEdgeMeta> e : graph.getSuccessorEdges(set, ZClassEdgeFilter.create(filter), transitive)) {
            ret.add(new ZClassEdge(e));
        }
        return ret;
    }

    public Set<ZIClassEdge> getPredecessorEdges(Set<String> set, ZIClassEdgeFilter filter, boolean transitive)
        throws Exception {
        Set<ZIClassEdge> ret = new HashSet<ZIClassEdge>();
        for (ZIEdge<String, ZIClassEdgeMeta> e : graph.getPredecessorEdges(set, ZClassEdgeFilter.create(filter), transitive)) {
            ret.add(new ZClassEdge(e));
        }
        return ret;
    }

    public Set<ZIClassEdge> getSuccessorEdges(String name, boolean transitive)
        throws Exception {
        return getSuccessorEdges(name, null, transitive);
    }

    public Set<ZIClassEdge> getEdges(ZIClassEdgeFilter filter)
        throws Exception {
        Set<ZIClassEdge> ret = new HashSet<ZIClassEdge>();
        for (ZIEdge<String, ZIClassEdgeMeta> e : graph.getEdgeSet(ZClassEdgeFilter.create(filter))) {
            ret.add(new ZClassEdge(e));
        }
        return ret;
    }

    public Set<String> getPredecessors(String name, boolean transitive)
        throws Exception {
        return graph.getPredecessors(name, null, transitive);
    }

    public Set<String> getSuccessors(String name, boolean transitive)
        throws Exception {
        return graph.getSuccessors(name, null, transitive);
    }

    public Set<String> getSuccessors(Set<String> set, ZIClassEdgeFilter filter, boolean transitive)
        throws Exception {
        return graph.getSuccessors(set, ZClassEdgeFilter.create(filter), transitive);
    }

    public Set<String> getSuccessors(String item, ZIClassEdgeFilter filter, boolean transitive)
        throws Exception {
        return graph.getSuccessors(item, ZClassEdgeFilter.create(filter), transitive);
    }

    public Set<String> getPredecessors(Set<String> set, ZIClassEdgeFilter filter, boolean transitive)
        throws Exception {
        return graph.getPredecessors(set, ZClassEdgeFilter.create(filter), transitive);
    }

    public Set<String> getPredecessors(String item, ZIClassEdgeFilter filter, boolean transitive)
        throws Exception {
        return graph.getPredecessors(item, ZClassEdgeFilter.create(filter), transitive);
    }

    public ZIPath getShortestPath(String v1, String v2) {
        List<ZIEdge<String, ZIClassEdgeMeta>> l = graph.getShortestPath(v1, v2);
        List<ZIClassEdge> ret = new ArrayList<ZIClassEdge>();
        if (l != null) {
            for (ZIEdge<String, ZIClassEdgeMeta> e : l) {
                ret.add(new ZClassEdge(e));
            }
        }
        return new ZPath(ret);
    }

    public boolean isSuccessor(String user, String used, boolean transitive)
        throws Exception {
        return graph.isSuccessor(user, used, null, transitive);
    }

    public boolean isSuccessor(String user, String used, ZIClassEdgeFilter filter, boolean transitive)
        throws Exception {
        return graph.isSuccessor(user, used, ZClassEdgeFilter.create(filter), transitive);
    }

    public boolean isPredecessor(String user, String used, boolean transitive)
        throws Exception {
        return graph.isPredecessor(user, used, null, transitive);
    }

    public boolean isPredecessor(String user, String used, ZIClassEdgeFilter filter, boolean transitive)
        throws Exception {
        return graph.isPredecessor(user, used, ZClassEdgeFilter.create(filter), transitive);
    }

    public ZIClassEdge getEdge(String t1, String t2) {
        ZIEdge<String, ZIClassEdgeMeta> e = graph.getEdge(t1, t2);
        ZIClassEdge ret = new ZClassEdge(e);
        return ret;
    }

    public String toXML()
        throws Exception {
        return ZXMLUtil.toXML(toElement());
    }

    public Element toElement()
        throws Exception {

        List<String> nodes = new ArrayList<String>(getClasses());
        List<ZIClassEdge> edges = new ArrayList<ZIClassEdge>(getEdges());
        Collections.sort(nodes);
        Collections.sort(edges);

        Element ret = new Element("classinfo");

        for (String node : nodes) {
            Element nodeElem = new Element("class");
            nodeElem.setAttribute("id", node);
            nodeElem.addContent(getClassMeta(node).toElement());
            ret.addContent(nodeElem);
        }

        for (ZIClassEdge edge : edges) {
            ret.addContent(edge.toElement());
        }
        return ret;
    }

    public ZIGraph<String, ZIClassEdgeMeta, ZIClassMeta> getGraph() {
        return graph;
    }

    public ZLocation getOrCreateLocation(String name, ZIManifest manifest) {
        ZLocation ret = locations.get(name);
        if (ret == null) {
            ret = new ZLocation(name, manifest);
            locations.put(name, ret);
        }
        return ret;
        //        return new ZLocation(name, manifest);
    }

    public boolean isClass(String name)
        throws Exception {
        return graph.isNode(name);
    }

    public boolean isEdge(String name1, String name2)
        throws Exception {
        return graph.isEdge(name1, name2);
    }

}
