/*
 * Copyright 2008 Gerd Ziegler (www.gerdziegler.de) Licensed under the Apache
 * License, Version 2.0 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.apache.org/licenses/LICENSE-2.0 Unless required by applicable law
 * or agreed to in writing, software distributed under the License is
 * distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the specific language
 * governing permissions and limitations under the License. @author
 * www.gerdziegler.de
 */
package org.ztest.classinfo.javassist;

import java.io.DataInputStream;
import java.io.File;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.jar.Attributes;
import java.util.jar.Manifest;

import javassist.bytecode.AnnotationsAttribute;
import javassist.bytecode.ClassFile;
import javassist.bytecode.ConstPool;
import javassist.bytecode.Descriptor;
import javassist.bytecode.FieldInfo;
import javassist.bytecode.MethodInfo;
import javassist.bytecode.annotation.Annotation;

import org.apache.log4j.Logger;
import org.zclasspath.ZClassPathClass;
import org.zclasspath.ZClassPathScanner;
import org.zclasspath.ZIClassPathElement;
import org.zclasspath.ZIClassPathLocation;
import org.zclasspath.ZIClassPathVisitor;
import org.ztest.classinfo.ZIClassEdge;
import org.ztest.classinfo.ZIClassInfo;
import org.ztest.classinfo.ZIClassInfoFactory;
import org.ztest.classinfo.ZIClassMeta;
import org.ztest.classinfo.ZILocation;
import org.ztest.classinfo.impl.ZClassEdgeMeta;
import org.ztest.classinfo.impl.ZClassInfo;
import org.ztest.classinfo.impl.ZField;
import org.ztest.classinfo.impl.ZManifest;
import org.ztest.classinfo.impl.ZMethod;
import org.ztest.classinfo.ser.ZClassInfoSerReader;
import org.ztest.classinfo.ser.ZClassInfoSerWriter;

public class ZClassInfoFactory implements ZIClassInfoFactory {
    private Logger log = Logger.getLogger(ZClassInfoFactory.class);

    private final ZClassPathScanner scanner;

    private ZIClassInfoFactoryListener listener;

    private boolean cache = false;

    public ZClassInfoFactory(ZClassPathScanner scanner) {
        this(scanner, false);
    }

    public ZClassInfoFactory(ZClassPathScanner scanner, boolean cache) {
        this.scanner = scanner;
        this.cache = cache;
    }

    public ZIClassInfo create()
        throws Exception {
        File file = null;
        if (cache) {
            long hash = computeHash(scanner);
            File tmpDir = new File(System.getProperty("java.io.tmpdir"));
            file = new File(tmpDir, "ztest-" + hash + ".zip");
            if (file.exists()) {
                try {
                    log.info("reading from " + file.getAbsolutePath());
                    ZIClassInfo ret = new ZClassInfoSerReader(file).read();
                    return ret;
                } catch (Exception e) {
                    log.error(file.getAbsolutePath(), e);
                    file.delete();
                }
            }
        }

        ZIClassInfo ret = createImpl();
        if (cache) {
            try {
                log.info("writing to " + file.getAbsolutePath());
                new ZClassInfoSerWriter(file).write(ret);
            } catch (Exception e) {
                log.error(file.getAbsolutePath(), e);
                log.info("will delete cache file on exit: " + file.getAbsolutePath());
                file.deleteOnExit();
            }
        }

        return ret;
    }

    private ZIClassInfo createImpl()
        throws Exception {
        final ZClassInfo classInfo = new ZClassInfo();
        scanner.setVisitor(new ZIClassPathVisitor() {
            public void visitClass(ZClassPathClass cpc)
                throws Exception {
                ZManifest man = new ZManifest();
                Manifest manifest = cpc.getClassPathElement().getManifest();
                if (manifest != null) {
                    for (Object key : manifest.getMainAttributes().keySet()) {
                        String value = manifest.getMainAttributes().getValue(key.toString());
                        man.getMainAttributes().put(key.toString(), value);
                    }

                    for (String key : manifest.getEntries().keySet()) {
                        Map<String, String> value = new HashMap<String, String>();
                        man.getEntries().put(key, value);
                        Attributes attr = manifest.getEntries().get(key);
                        for (Object attrKey : attr.keySet()) {
                            value.put(attrKey.toString(), attr.getValue(attrKey.toString()));
                        }
                    }
                }

                ZIClassPathLocation classPathLocation = cpc.getClassPathElement().getClassPathLocation();
                StringBuffer locationName = new StringBuffer(classPathLocation.getClassPathItem().getName());
                for(String parent: classPathLocation.getParentNames()){
                    locationName.append('/');
                    locationName.append(parent);
                }               
                
                ZILocation location = classInfo.getOrCreateLocation(locationName.toString(), man);

                DataInputStream in = new DataInputStream(cpc.getClassPathElement().getInputStream());
                try {
                    ClassFile cc = new ClassFile(in);
                    processClass(classInfo, cc, location);
                    if (listener != null) {
                        listener.processing(classInfo, cc);
                    }
                } finally {
                    in.close();
                }
            }

            public void visitResource(ZIClassPathElement arg0)
                throws Exception {
            }
        });
        long startTime = System.currentTimeMillis();
        log.info("begin scan");
        scanner.scan();

        long time = (System.currentTimeMillis() - startTime) / 1000;
        log.info("end scan [" + time + " s]");

        if (classInfo.getClasses().isEmpty()) {
            throw new Exception("no classes found");
        }

        return classInfo;
    }

    private long computeHash(ZClassPathScanner scanner)
        throws Exception {

        class MyClassPathVisitor implements ZIClassPathVisitor {
            long ret = 0;

            public void visitClass(ZClassPathClass cpc)
                throws Exception {
                ret += cpc.getClassPathElement().getClassPathLocation().getClassPathItem().getHash();
            }

            public void visitResource(ZIClassPathElement arg0)
                throws Exception {
            }
        }

        MyClassPathVisitor vis = new MyClassPathVisitor();
        scanner.setVisitor(vis);
        scanner.scan();
        return vis.ret;
    }

    private void processClass(final ZIClassInfo classInfo, ClassFile cc, ZILocation location)
        throws Exception {
        String className = cc.getName();
        classInfo.addClass(className);

        ZIClassMeta cm = classInfo.getClassMeta(className);
        cm.getLocations().add(location);
        cm.setModifiers(cc.getAccessFlags());

        AnnotationsAttribute attr = (AnnotationsAttribute)cc.getAttribute(AnnotationsAttribute.visibleTag);
        if (attr != null) {
            Annotation[] annotations = attr.getAnnotations();
            if (annotations.length > 0) {
                for (Annotation an : annotations) {
                    String annotationTypeName = an.getTypeName();
                    cm.getAnnotations().add(annotationTypeName);
                    log.debug("annotation " + annotationTypeName);
                }
            }
        }
        ZIClassEdge edge = classInfo.addEdgeAndNode(className, cc.getSuperclass());
        ((ZClassEdgeMeta)edge.getMetaData()).setUsedInExtends(true);

        scanInheritance(classInfo, cc.getSuperclass());
        log.debug(className + " -extends- " + cc.getSuperclass());

        for (String ifc : cc.getInterfaces()) {
            ZIClassEdge edge1 = classInfo.addEdgeAndNode(className, ifc);
            ((ZClassEdgeMeta)edge1.getMetaData()).setUsedInImplements(true);
            scanInheritance(classInfo, ifc);
            log.debug(className + " -implements- " + ifc);
        }

        ConstPool con = cc.getConstPool();
        Set<String> poolClassNames = con.getClassNames();
        for (String poolClassName : poolClassNames) {
            // String poolClassName = Descriptor.toClassName(name2);
            poolClassName = poolClassName.replace('/', '.');
            if (poolClassName.startsWith("[L") && poolClassName.endsWith(";")) {
                poolClassName = poolClassName.substring(2, poolClassName.length() - 1);
            }

            ZIClassEdge edge2 = classInfo.addEdgeAndNode(className, poolClassName);
            ((ZClassEdgeMeta)edge2.getMetaData()).setUsedInCode(true);

            scanInheritance(classInfo, poolClassName);
        }

        List<FieldInfo> fields = cc.getFields();
        for (FieldInfo field : fields) {
            String type = Descriptor.toClassName(field.getDescriptor());
            type = trimArray(type);
            int accessFlags = field.getAccessFlags();
            String name = field.getName();

            ZField fm = new ZField(name, accessFlags);
            ZIClassEdge edge1 = classInfo.addEdgeAndNode(className, type);
            edge1.getMetaData().getUsedInField().add(fm);

            scanInheritance(classInfo, type);
        }

        List<MethodInfo> methods = cc.getMethods();
        for (MethodInfo method : methods) {

            String desc = method.getDescriptor();
            String paramDesc = Descriptor.getParamDescriptor(desc);
            String returnValues = desc.substring(paramDesc.length());
            String returnType = Descriptor.toClassName(returnValues);

            returnType = trimArray(returnType);

            ZMethod zmethod = new ZMethod(method.getName(), method.getAccessFlags(), desc);
            cm.getMethods().add(zmethod);

            ZIClassEdge edge1 = classInfo.addEdgeAndNode(className, returnType);
            edge1.getMetaData().getUsedInMethod().add(zmethod);

            scanInheritance(classInfo, returnType);

            int begin = 0;
            boolean wasParam = false;
            Descriptor.Iterator iter = new Descriptor.Iterator(desc);
            while (iter.hasNext()) {
                int next = iter.next();
                if (iter.isParameter()) {
                    if (wasParam) {
                        String param = desc.substring(begin, next);
                        param = Descriptor.toClassName(param);
                        param = trimArray(param);

                        ZIClassEdge edge2 = classInfo.addEdgeAndNode(className, param);
                        edge2.getMetaData().getUsedInMethod().add(zmethod);

                        scanInheritance(classInfo, param);
                    }
                    wasParam = true;
                    begin = next;
                } else if (wasParam) {
                    String param = desc.substring(begin, next - 1);
                    param = Descriptor.toClassName(param);
                    param = trimArray(param);
                    ZIClassEdge edge2 = classInfo.addEdgeAndNode(className, param);
                    edge2.getMetaData().getUsedInMethod().add(zmethod);
                    scanInheritance(classInfo, param);
                    wasParam = false;
                }
            }
        }
    }

    // private void createEdge(final ZClassInfo classInfo, String c1, String c2,
    // ZIMeta meta) {
    // ZIClassEdge edge = classInfo.addEdgeAndNode(c1, c2);
    // edge.getMetaData().put(meta.getClass().getName(), meta);
    // }

    private String trimArray(String param) {
        if (param.endsWith("[]")) {
            return param.substring(0, param.length() - 2);
        } else {
            return param;
        }
    }

    private void scanInheritance(final ZIClassInfo classInfo, String className) {
        try {
            Class c = Class.forName(className);
            if (Object.class.equals(c)) {
                return;
            }

            ZIClassMeta cm = classInfo.getClassMeta(className);
            cm.setModifiers(c.getModifiers());

            for (Class ifc : c.getInterfaces()) {
                ZIClassEdge edge2 = classInfo.addEdgeAndNode(className, ifc.getName());
                ((ZClassEdgeMeta)edge2.getMetaData()).setUsedInImplements(true);
                log.debug(className + " -implements- " + ifc.getName());
                scanInheritance(classInfo, ifc.getName());
            }

            Class superClass = c.getSuperclass();
            ZIClassEdge edge2 = classInfo.addEdgeAndNode(className, superClass.getName());
            ((ZClassEdgeMeta)edge2.getMetaData()).setUsedInExtends(true);
            scanInheritance(classInfo, superClass.getName());
        } catch (Exception e) {
            log.debug("to allow full inheritance scan add to classpath: " + className);
        }
    }

    // public boolean isCache() {
    // return cache;
    // }
    //
    // public void setCache(boolean cache) {
    // this.cache = cache;
    // }

    public ZIClassInfoFactoryListener getListener() {
        return listener;
    }

    public void setListener(ZIClassInfoFactoryListener visitor) {
        this.listener = visitor;
    }

    public boolean isCache() {
        return cache;
    }

    public void setCache(boolean cache) {
        this.cache = cache;
    }
}
