/*
 * Copyright 2008 Gerd Ziegler (www.gerdziegler.de) Licensed under the Apache
 * License, Version 2.0 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.apache.org/licenses/LICENSE-2.0 Unless required by applicable law
 * or agreed to in writing, software distributed under the License is
 * distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the specific language
 * governing permissions and limitations under the License. @author
 * www.gerdziegler.de
 */
package org.ztest.classinfo.xml;

import java.io.BufferedReader;
import java.io.Reader;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.log4j.Logger;
import org.jdom.Document;
import org.jdom.Element;
import org.jdom.input.SAXBuilder;
import org.ztest.classinfo.ZIClassEdge;
import org.ztest.classinfo.ZIClassInfo;
import org.ztest.classinfo.impl.ZClassEdgeMeta;
import org.ztest.classinfo.impl.ZClassInfo;
import org.ztest.classinfo.impl.ZClassMeta;
import org.ztest.classinfo.impl.ZField;
import org.ztest.classinfo.impl.ZManifest;
import org.ztest.classinfo.impl.ZMethod;

public class ZClassInfoXMLReader {
    private Logger log = Logger.getLogger(ZClassInfoXMLReader.class);

    private final Reader reader;

    public ZClassInfoXMLReader(Reader reader) {
        this.reader = reader;
    }

    public ZIClassInfo read()
        throws Exception {
        ZClassInfo classInfo = new ZClassInfo();

        SAXBuilder builder = new SAXBuilder();
        BufferedReader buff = new BufferedReader(reader, 16 * 1024);
        Document doc = builder.build(reader);
        buff.close();
        readGraph(classInfo, doc.getRootElement());

        return classInfo;
    }

    private void readGraph(ZClassInfo classInfo, Element graph)
        throws Exception {
        for (Element node : (List<Element>)graph.getChildren("class")) {
            readClass(classInfo, node);
        }
        for (Element node : (List<Element>)graph.getChildren("edge")) {
            readEdge(classInfo, node);
        }
    }

    private void readClass(ZClassInfo classInfo, Element node)
        throws Exception {
        String name = node.getChildText("name");
        classInfo.addClass(name);
        ZClassMeta meta = (ZClassMeta)classInfo.getClassMeta(name);
        readClassMeta(classInfo, node, meta);
    }

    private void readClassMeta(ZClassInfo classInfo, Element parent, ZClassMeta meta) {
        meta.setModifiers(Integer.parseInt(parent.getChildText("modifiers"), 2));

        for (Element node : (List<Element>)parent.getChildren("location")) {
            ZManifest man = readManifest(node);
            String loc = node.getChildText("name");
            meta.getLocations().add(classInfo.getOrCreateLocation(loc, man));
        }

        for (Element node : (List<Element>)parent.getChildren("annotation")) {
            String txt = node.getText();
            meta.getAnnotations().add(txt);
        }

        meta.getFields().addAll(readFields(parent));
        meta.getMethods().addAll(readMethods(parent));
    }

    private ZManifest readManifest(Element parent) {
        Element manNode = parent.getChild("manifest");
        if (manNode == null) {
            return null;
        }

        ZManifest ret = new ZManifest();

        Element mainAttributesNode = manNode.getChild("main-attributes");
        for (Element propNode : (List<Element>)mainAttributesNode.getChildren("prop")) {
            String key = propNode.getAttributeValue("name");
            String value = propNode.getText();
            ret.getMainAttributes().put(key, value);
        }

        for (Element entryNode : (List<Element>)manNode.getChildren("entry")) {
            String name = entryNode.getAttributeValue("name");
            Map<String, String> attr = new HashMap<String, String>();
            ret.getEntries().put(name, attr);
            for (Element propNode : (List<Element>)entryNode.getChildren("prop")) {
                String key = propNode.getAttributeValue("name");
                String value = propNode.getText();
                attr.put(key, value);
            }
        }
        return ret;
    }

    private List<ZMethod> readMethods(Element parent) {
        List<ZMethod> ret = new ArrayList<ZMethod>();
        for (Element node : (List<Element>)parent.getChildren("method")) {
            String name = node.getChildText("name");
            int modifiers = Integer.parseInt(node.getChildText("modifiers"), 2);
            String descriptor = node.getChildText("descriptor");
            ret.add(new ZMethod(name, modifiers, descriptor));
        }
        return ret;
    }

    private List<ZField> readFields(Element parent) {
        List<ZField> ret = new ArrayList<ZField>();
        for (Element node : (List<Element>)parent.getChildren("field")) {
            String name = node.getChildText("name");
            int modifiers = Integer.parseInt(node.getChildText("modifiers"), 2);
            ret.add(new ZField(name, modifiers));
        }
        return ret;
    }

    private void readEdge(ZIClassInfo classInfo, Element parent)
        throws Exception {
        String src = parent.getChildText("source");
        String tgt = parent.getChildText("target");
        ZIClassEdge edge = classInfo.addEdge(src, tgt);

        ZClassEdgeMeta meta = (ZClassEdgeMeta)edge.getMetaData();
        meta.setUsedInCode(Boolean.parseBoolean(parent.getChildText("usedInCode")));
        meta.setUsedInExtends(Boolean.parseBoolean(parent.getChildText("usedInExtends")));
        meta.setUsedInImplements(Boolean.parseBoolean(parent.getChildText("usedInImplements")));

        meta.getUsedInField().addAll(readFields(parent));
        meta.getUsedInMethod().addAll(readMethods(parent));
    }

}
