/*
 * Copyright 2008 Gerd Ziegler (www.gerdziegler.de) Licensed under the Apache
 * License, Version 2.0 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.apache.org/licenses/LICENSE-2.0 Unless required by applicable law
 * or agreed to in writing, software distributed under the License is
 * distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the specific language
 * governing permissions and limitations under the License. @author
 * www.gerdziegler.de
 */
package org.ztest.classinfo.xml;

import java.io.BufferedWriter;
import java.io.Writer;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.Map;

import org.apache.log4j.Logger;
import org.jdom.Document;
import org.jdom.Element;
import org.jdom.Text;
import org.jdom.output.Format;
import org.jdom.output.XMLOutputter;
import org.ztest.classinfo.ZIClassEdge;
import org.ztest.classinfo.ZIClassEdgeMeta;
import org.ztest.classinfo.ZIClassInfo;
import org.ztest.classinfo.ZIClassMeta;
import org.ztest.classinfo.ZIField;
import org.ztest.classinfo.ZILocation;
import org.ztest.classinfo.ZIManifest;
import org.ztest.classinfo.ZIMethod;

public class ZClassInfoXMLWriter {
    private Logger log = Logger.getLogger(ZClassInfoXMLWriter.class);

    private final Writer writer;

    public ZClassInfoXMLWriter(Writer writer) {
        this.writer = writer;
    }

    public void write(ZIClassInfo classInfo)
        throws Exception {

        Element root = toElement(classInfo);

        Document doc = new Document(root);
        XMLOutputter outputter = new XMLOutputter(Format.getPrettyFormat().setExpandEmptyElements(false));

        BufferedWriter buff = new BufferedWriter(writer, 16 * 1024);
        outputter.output(doc, buff);
        buff.flush();
    }

    private Element toElement(ZIClassInfo classInfo)
        throws Exception {
        List<String> nodes = new ArrayList<String>(classInfo.getClasses());
        List<ZIClassEdge> edges = new ArrayList<ZIClassEdge>(classInfo.getEdges());
        Collections.sort(nodes);
        Collections.sort(edges);

        Element ret = new Element("ztest");

        for (String node : nodes) {
            ret.addContent(toElement(classInfo, node));
        }

        for (ZIClassEdge edge : edges) {
            ret.addContent(toElement(edge));
        }

        return ret;
    }

    private Element toElement(ZIClassInfo classInfo, String name)
        throws Exception {
        Element ret = new Element("class");
        ret.addContent(createTextElement("name", name));

        addElement(ret, classInfo.getClassMeta(name));

        return ret;
    }

    private Element toElement(ZIClassEdge edge)
        throws Exception {
        Element ret = new Element("edge");
        ret.addContent(createTextElement("source", edge.getSource()));
        ret.addContent(createTextElement("target", edge.getTarget()));
        addElement(ret, edge.getMetaData());
        return ret;
    }

    private Element addElement(Element ret, ZIClassMeta meta)
        throws Exception {
        ret.addContent(createTextElement("modifiers", Integer.toBinaryString(meta.getModifiers())));

        List<ZILocation> locations = meta.getLocations();
        Collections.sort(locations);
        for (ZILocation loc : locations) {
            Element locNode = new Element("location");
            locNode.addContent(createTextElement("name", loc.getName()));
            if (loc.getManifest() != null) {
                locNode.addContent(toElement(loc.getManifest()));
            }
            ret.addContent(locNode);
        }

        List<String> annotations = meta.getAnnotations();
        Collections.sort(annotations);
        for (String ann : annotations) {
            ret.addContent(createTextElement("annotation", ann));
        }

        addFields(ret, meta.getFields());
        addMethods(ret, meta.getMethods());

        return ret;
    }

    private Element toElement(ZIManifest man) {
        Element ret = new Element("manifest");
        
        List<String> mainKeys = new ArrayList<String>(man.getMainAttributes().keySet());
        Collections.sort(mainKeys);
        Element mainAttributesElement = new Element("main-attributes");
        ret.addContent(mainAttributesElement);
        for(String s: mainKeys){
            String val = man.getMainAttributes().get(s);
            Element valNode = createTextElement("prop", val);
            valNode.setAttribute("name", s);
            mainAttributesElement.addContent(valNode);
        }        
        
        List<String> keys = new ArrayList<String>(man.getEntries().keySet());
        Collections.sort(keys);
        for (String key : keys) {
            Element entry = new Element("entry");
            entry.setAttribute("name", key);
            ret.addContent(entry);
            Map<String, String> attr = man.getEntries().get(key);

            List<String> attrNames = new ArrayList<String>();
            for (Object s : attr.keySet()) {
                attrNames.add(s.toString());
            }
            Collections.sort(attrNames);
            for (String s : attrNames) {
                String val = attr.get(s);
                Element valNode = createTextElement("prop", val);
                valNode.setAttribute("name", s);
                entry.addContent(valNode);
            }
        }
        return ret;
    }

    private void addFields(Element parent, Collection<ZIField> fieldSet) {
        List<ZIField> fields = new ArrayList<ZIField>(fieldSet);
        Collections.sort(fields);
        for (ZIField m : fields) {
            Element fieldNode = new Element("field");
            fieldNode.addContent(createTextElement("name", m.getName()));
            fieldNode.addContent(createTextElement("modifiers", Integer.toBinaryString(m.getModifiers())));
            parent.addContent(fieldNode);
        }
    }

    private void addMethods(Element parent, Collection<ZIMethod> methodSet) {
        List<ZIMethod> methods = new ArrayList<ZIMethod>(methodSet);
        Collections.sort(methods);
        for (ZIMethod m : methods) {
            Element fieldNode = new Element("method");
            fieldNode.addContent(createTextElement("name", m.getName()));
            fieldNode.addContent(createTextElement("modifiers", Integer.toBinaryString(m.getModifiers())));
            fieldNode.addContent(createTextElement("descriptor", m.getDescriptor()));
            parent.addContent(fieldNode);
        }
    }

    private Element createTextElement(String name, String text) {
        Element ret = new Element(name);
        ret.addContent(new Text(text));
        return ret;
    }

    private void addElement(Element ret, ZIClassEdgeMeta meta) {
        ret.addContent(createTextElement("usedInExtends", String.valueOf(meta.isUsedInExtends())));
        ret.addContent(createTextElement("usedInImplements", String.valueOf(meta.isUsedInImplements())));
        ret.addContent(createTextElement("usedInCode", String.valueOf(meta.isUsedInCode())));
        addFields(ret, meta.getUsedInField());
        addMethods(ret, meta.getUsedInMethod());
    }
}
