/*
 * Copyright 2008 Gerd Ziegler (www.gerdziegler.de) Licensed under the Apache
 * License, Version 2.0 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.apache.org/licenses/LICENSE-2.0 Unless required by applicable law
 * or agreed to in writing, software distributed under the License is
 * distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the specific language
 * governing permissions and limitations under the License. @author
 * www.gerdziegler.de
 */
package org.ztest.graph.jgrapht;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.log4j.Logger;
import org.jdom.Element;
import org.jgrapht.EdgeFactory;
import org.jgrapht.alg.DijkstraShortestPath;
import org.jgrapht.alg.StrongConnectivityInspector;
import org.jgrapht.graph.DefaultDirectedGraph;
import org.ztest.graph.ZIEdge;
import org.ztest.graph.ZIEdgeFilter;
import org.ztest.graph.ZIGraph;
import org.ztest.graph.ZINodeFilter;
import org.ztest.jdom.ZWriteXML;
import org.ztest.jdom.ZXMLUtil;

public class ZJGraphTGraph<T, EM extends ZWriteXML, NM extends ZWriteXML> implements ZIGraph<T, EM, NM> {
    private static final Logger log = Logger.getLogger(ZJGraphTGraph.class);

    private final Map<T, Map<String, NM>> nodeMetaMap = new HashMap<T, Map<String, NM>>();

    class MyEdgeFactory implements Serializable, EdgeFactory<T, ZIEdge<T, EM>> {
        public ZIEdge<T, EM> createEdge(T n1, T n2) {
            return new ZEdge<T, EM>(n1, n2);
        }
    };

    private MyEdgeFactory edgeFactory = new MyEdgeFactory();

    private DefaultDirectedGraph<T, ZIEdge<T, EM>> graph = new DefaultDirectedGraph<T, ZIEdge<T, EM>>(edgeFactory) {
        @Override
        public boolean isAllowingLoops() {
            return false;
        }

        @Override
        public boolean isAllowingMultipleEdges() {
            return false;
        }
    };

    public ZIEdge<T, EM> addEdge(T n1, T n2) {
        if (!isNode(n1)) {
            addNode(n1);
        }
        if (!isNode(n2)) {
            addNode(n2);
        }
        ZIEdge<T, EM> edge = graph.getEdge(n1, n2);
        if (edge == null) {
            edge = graph.addEdge(n1, n2);
        }
        return edge;
    }

    public void addNode(T n) {
        graph.addVertex(n);
        nodeMetaMap.put(n, new HashMap<String, NM>());
    }

    public Set<ZIEdge<T, EM>> getEdgeSet() {
        return graph.edgeSet();
    }

    public Set<T> getNodeSet() {
        Set<T> ret = graph.vertexSet();
        if (ret == null) {
            ret = new HashSet<T>();
        }
        return ret;
    }

    public Set<T> getNodeSet(ZINodeFilter<T> filter)
        throws Exception {
        Set<T> set = graph.vertexSet();

        Set<T> ret = new HashSet<T>();
        if (set != null) {
            for (T t : set) {
                if (filter == null || filter.accept(t)) {
                    ret.add(t);
                }
            }
        }
        return ret;
    }

    public List<ZIEdge<T, EM>> getShortestPath(T n1, T n2) {
        List<ZIEdge<T, EM>> ret = DijkstraShortestPath.findPathBetween(graph, n1, n2);
        return ret;
    }

    public List<Set<T>> getCycles(boolean includeSelfcycles) {
        StrongConnectivityInspector<T, ZIEdge<T, EM>> conn = new StrongConnectivityInspector<T, ZIEdge<T, EM>>(graph);
        List<Set<T>> cycles = conn.stronglyConnectedSets();
        if (includeSelfcycles) {
            return cycles;
        }

        List<Set<T>> ret = new ArrayList<Set<T>>();
        for (Set<T> cycle : cycles) {
            if (cycle.size() > 1) {
                ret.add(cycle);
            }
        }
        return ret;
    }

    public List<Set<T>> getCycles() {
        return getCycles(false);
    }

    public boolean isEdge(T n1, T n2) {
        return graph.containsEdge(n1, n2);
    }

    public boolean isNode(T n) {
        return graph.containsVertex(n);
    }

    public Set<T> getSuccessors(T t1, ZIEdgeFilter<T, EM> filter, boolean transitive)
        throws Exception {
        Set<T> visit = new HashSet<T>();
        visit.add(t1);
        return getSuccessors(visit, filter, transitive);
    }

    public Set<T> getSuccessors(Set<T> itemSet, ZIEdgeFilter<T, EM> filter, boolean transitive)
        throws Exception {
        Set<ZIEdge<T, EM>> edges = getSuccessorEdges(itemSet, filter, transitive);
        Set<T> ret = new HashSet<T>();
        for (ZIEdge<T, EM> e : edges) {
            ret.add(e.getSource());
            ret.add(e.getTarget());
        }
        ret.removeAll(itemSet);
        return ret;
    }

    public Set<ZIEdge<T, EM>> getSuccessorEdges(T t1, ZIEdgeFilter<T, EM> filter, boolean transitive)
        throws Exception {
        Set<T> visit = new HashSet<T>();
        visit.add(t1);
        return getSuccessorEdges(visit, filter, transitive);
    }

    public Set<ZIEdge<T, EM>> getSuccessorEdges(Set<T> itemSet, ZIEdgeFilter<T, EM> filter, boolean transitive)
        throws Exception {
        Set<ZIEdge<T, EM>> ret = new HashSet<ZIEdge<T, EM>>();

        Set<T> visit = new HashSet<T>();
        visit.addAll(itemSet);
        Set<T> visited = new HashSet<T>();
        while (!visit.isEmpty()) {
            T crt = visit.iterator().next();
            visit.remove(crt);
            visited.add(crt);
            if (!isNode(crt)) {
                continue;
            }
            Set<ZIEdge<T, EM>> edges = graph.outgoingEdgesOf(crt);
            for (ZIEdge<T, EM> e : edges) {
                T next1 = e.getTarget();
                if (filter == null || filter.accept(e)) {
                    ret.add(e);

                    T next = e.getTarget();
                    if (crt.equals(next) || itemSet.contains(next)) {
                        continue;
                    }
                    if (!visited.contains(next)) {
                        if (transitive) {
                            visit.add(next);
                        } else {
                            visited.add(next);
                        }
                    }
                }
            }
        }
        return ret;
    }

    public Set<T> getPredecessors(T t1, ZIEdgeFilter<T, EM> filter, boolean transitive)
        throws Exception {
        Set<T> visit = new HashSet<T>();
        visit.add(t1);
        return getPredecessors(visit, filter, transitive);
    }

    public Set<T> getPredecessors(Set<T> itemSet, ZIEdgeFilter<T, EM> filter, boolean transitive)
        throws Exception {
        Set<ZIEdge<T, EM>> edges = getPredecessorEdges(itemSet, filter, transitive);
        Set<T> ret = new HashSet<T>();
        for (ZIEdge<T, EM> e : edges) {
            ret.add(e.getSource());
            ret.add(e.getTarget());
        }
        ret.removeAll(itemSet);
        return ret;
    }

    public Set<ZIEdge<T, EM>> getPredecessorEdges(T t1, ZIEdgeFilter<T, EM> filter, boolean transitive)
        throws Exception {
        Set<T> visit = new HashSet<T>();
        visit.add(t1);
        return getPredecessorEdges(visit, filter, transitive);
    }

    public Set<ZIEdge<T, EM>> getPredecessorEdges(Set<T> itemSet, ZIEdgeFilter<T, EM> filter, boolean transitive)
        throws Exception {
        Set<ZIEdge<T, EM>> ret = new HashSet<ZIEdge<T, EM>>();

        Set<T> visit = new HashSet<T>();
        visit.addAll(itemSet);
        Set<T> visited = new HashSet<T>();
        while (!visit.isEmpty()) {
            T crt = visit.iterator().next();
            visit.remove(crt);
            visited.add(crt);
            if (!isNode(crt)) {
                continue;
            }
            Set<ZIEdge<T, EM>> edges = graph.incomingEdgesOf(crt);
            for (ZIEdge<T, EM> e : edges) {
                if (filter == null || filter.accept(e)) {
                    ret.add(e);

                    T next = e.getSource();
                    if (crt.equals(next) || itemSet.contains(next)) {
                        continue;
                    }
                    if (!visited.contains(next)) {
                        if (transitive) {
                            visit.add(next);
                        } else {
                            visited.add(next);
                        }
                    }
                }
            }
        }
        return ret;
    }

    public boolean isSuccessor(T t1, T t2, ZIEdgeFilter<T, EM> filter, boolean transitive)
        throws Exception {
        Set<T> visited = new HashSet<T>();
        Set<T> visit = new HashSet<T>();
        visit.add(t1);
        while (!visit.isEmpty()) {
            T crt = visit.iterator().next();
            visit.remove(crt);
            visited.add(crt);
            if (!isNode(crt)) {
                continue;
            }
            Set<ZIEdge<T, EM>> edges = graph.outgoingEdgesOf(crt);
            for (ZIEdge<T, EM> e : edges) {
                T next = e.getTarget();
                if (crt.equals(next) || t1.equals(next)) {
                    continue;
                }
                if (filter == null || filter.accept(e)) {
                    if (next.equals(t2)) {
                        return true;
                    } else if (transitive && !visited.contains(next)) {
                        visit.add(next);
                    }
                }
            }
        }

        return false;
    }

    public boolean isPredecessor(T t1, T t2, ZIEdgeFilter<T, EM> filter, boolean transitive)
        throws Exception {
        Set<T> visited = new HashSet<T>();
        Set<T> visit = new HashSet<T>();
        visit.add(t1);
        while (!visit.isEmpty()) {
            T crt = visit.iterator().next();
            visit.remove(crt);
            visited.add(crt);
            if (!isNode(crt)) {
                continue;
            }
            Set<ZIEdge<T, EM>> edges = graph.incomingEdgesOf(crt);
            for (ZIEdge<T, EM> e : edges) {
                T next = e.getSource();
                if (crt.equals(next) || t1.equals(next)) {
                    continue;
                }
                if (filter == null || filter.accept(e)) {
                    if (next.equals(t2)) {
                        return true;
                    } else if (transitive && !visited.contains(next)) {
                        visit.add(next);
                    }
                }
            }
        }

        return false;
    }

    public Map<String, NM> getNodeMeta(T n) {
        return nodeMetaMap.get(n);
    }

    public DefaultDirectedGraph<T, ZIEdge<T, EM>> getGraph() {
        return graph;
    }

    public ZIEdge<T, EM> getEdge(T t1, T t2) {
        return graph.getEdge(t1, t2);
    }

    public void fromElement(Element object) {
        throw new RuntimeException("not implemented");
    }

    public Element toElement()
        throws Exception {
        Element ret = new Element("graph");
        Element nodes = new Element("nodes");
        ret.addContent(nodes);
        List<T> sorted = new ArrayList<T>(graph.vertexSet());
        Collections.sort(sorted, new Comparator<T>() {
            public int compare(T o1, T o2) {
                return o1.toString().compareTo(o2.toString());
            }
        });
        for (T t : sorted) {
            Element nodeElem = new Element("node");
            nodeElem.setAttribute("name", t.toString());
            nodes.addContent(nodeElem);

            for (NM meta : getNodeMeta(t).values()) {
                nodeElem.addContent(meta.toElement());
            }
        }

        Element edges = new Element("edges");
        ret.addContent(edges);
        List<ZIEdge<T, EM>> sortedEdges = new ArrayList<ZIEdge<T, EM>>(graph.edgeSet());
        Collections.sort(sortedEdges, new Comparator<ZIEdge<T, EM>>() {
            public int compare(ZIEdge<T, EM> o1, ZIEdge<T, EM> o2) {
                int ret = o1.getSource().toString().compareTo(o2.getSource().toString());
                if (ret != 0) {
                    return ret;
                }
                return o1.getTarget().toString().compareTo(o2.getTarget().toString());
            }
        });
        for (ZIEdge<T, EM> e : sortedEdges) {
            edges.addContent(e.toElement());
        }
        return ret;
    }

    public String toXML()
        throws Exception {
        return ZXMLUtil.toXML(toElement());
    }

    public Set<ZIEdge<T, EM>> getEdgeSet(ZIEdgeFilter<T, EM> filter)
        throws Exception {
        Set<ZIEdge<T, EM>> ret = new HashSet<ZIEdge<T, EM>>();
        for (ZIEdge<T, EM> edge : graph.edgeSet()) {
            if (filter == null || filter.accept(edge)) {
                ret.add(edge);
            }
        }
        return ret;
    }
}
