/*
 * Copyright 2008 Gerd Ziegler (www.gerdziegler.de) Licensed under the Apache
 * License, Version 2.0 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.apache.org/licenses/LICENSE-2.0 Unless required by applicable law
 * or agreed to in writing, software distributed under the License is
 * distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the specific language
 * governing permissions and limitations under the License. @author
 * www.gerdziegler.de
 */
package org.ztest.output.classinfo;

import java.io.Writer;
import java.util.HashSet;
import java.util.Set;

import org.ztest.classinfo.ZIClassEdge;
import org.ztest.classinfo.ZIClassEdgeFilter;
import org.ztest.classinfo.ZIClassFilter;
import org.ztest.classinfo.ZIClassInfo;
import org.ztest.filter.ZClassEdgeFilterAcceptAll;
import org.ztest.filter.ZClassEdgeFilterByClassFilter;
import org.ztest.filter.ZClassEdgeFilterByInheritance;
import org.ztest.filter.ZClassFilterAcceptAll;
import org.ztest.output.graphml.ZGraphMLWriter;
import org.ztest.output.graphml.ZIGraphMLEdge;
import org.ztest.output.graphml.ZIGraphMLNode;

public class ZClassInfoWriter {
    private final ZIClassInfo classInfo;

    private ZIClassFilter classFilter = null;

    private ZIClassEdgeFilter classEdgeFilter = null;

    private ZClassInfoDecorator decorator;

    public static ZClassInfoWriter createClassHierarchyWriter(ZIClassInfo classInfo, ZIClassFilter classFilter)
        throws Exception {
        ZClassEdgeFilterByClassFilter classEdgeFilter = new ZClassEdgeFilterByClassFilter(classFilter);
        ZClassInfoWriter w = new ZClassInfoWriter(classInfo);
        w.setClassFilter(classFilter);
        w.setClassEdgeFilter(new ZClassEdgeFilterByInheritance(classEdgeFilter));
        return w;
    }

    public ZClassInfoWriter(final ZIClassInfo classInfo, ZIClassFilter classFilter, ZIClassEdgeFilter classEdgeFilter, ZClassInfoDecorator decorator) {
        super();
        this.classInfo = classInfo;
        this.classFilter = classFilter;
        this.classEdgeFilter = classEdgeFilter;
        this.decorator = decorator;
    }

    public ZClassInfoWriter(final ZIClassInfo classInfo, ZIClassFilter classFilter, ZIClassEdgeFilter classEdgeFilter) {
        super();
        this.classInfo = classInfo;
        this.classFilter = classFilter;
        this.classEdgeFilter = classEdgeFilter;
    }

    public ZClassInfoWriter(final ZIClassInfo classInfo, ZIClassFilter classFilter) {
        super();
        this.classInfo = classInfo;
        this.classFilter = classFilter;
    }

    public ZClassInfoWriter(final ZIClassInfo classInfo) {
        super();
        this.classInfo = classInfo;
        decorator = new ZClassInfoDecorator(classInfo);
    }

    public void write(Writer writer)
        throws Exception {
        if (decorator == null) {
            decorator = new ZClassInfoDecorator(classInfo);
        }
        if (classFilter == null) {
            classFilter = new ZClassFilterAcceptAll();
        }
        if (classEdgeFilter == null) {
            classEdgeFilter = new ZClassEdgeFilterAcceptAll();
        }

        Set<ZIGraphMLNode> nodes = new HashSet<ZIGraphMLNode>();
        for (String name : classInfo.getClasses(classFilter)) {
            nodes.add(createNode(name));
        }

        Set<ZIGraphMLEdge> edges = new HashSet<ZIGraphMLEdge>();
        for (ZIClassEdge edge : classInfo.getEdges(classEdgeFilter)) {
            edges.add(createEdge(edge));
        }

        ZGraphMLWriter exporter = new ZGraphMLWriter(nodes, edges, decorator);
        exporter.write(writer);
    }

    public static Set<ZIGraphMLEdge> createEdges(final Set<ZIClassEdge> edges) {
        Set<ZIGraphMLEdge> ret = new HashSet<ZIGraphMLEdge>();
        for(ZIClassEdge edge: edges) {
            ret.add(createEdge(edge));
        }
        return ret;
    }

    public static ZIGraphMLEdge createEdge(final ZIClassEdge edge) {
        ZIGraphMLEdge ret = new ZIGraphMLEdge() {
            public String getSourceId()
                throws Exception {
                return edge.getSource();
            }

            public String getTargetId()
                throws Exception {
                return edge.getTarget();
            }
        };
        return ret;
    }

    public static Set<ZIGraphMLNode> createNodes(final Set<String> names) {
        Set<ZIGraphMLNode> ret = new HashSet<ZIGraphMLNode>();
        for(String name: names) {
            ret.add(createNode(name));
        }
        return ret;
    }

    public static ZIGraphMLNode createNode(final String name) {
        ZIGraphMLNode ret = new ZIGraphMLNode() {
            public String getId() {
                return name;
            }
        };
        return ret;
    }

    public ZIClassEdgeFilter getClassEdgeFilter() {
        return classEdgeFilter;
    }

    public void setClassEdgeFilter(ZIClassEdgeFilter classEdgeFilter) {
        this.classEdgeFilter = classEdgeFilter;
    }

    public ZIClassFilter getClassFilter() {
        return classFilter;
    }

    public void setClassFilter(ZIClassFilter classFilter) {
        this.classFilter = classFilter;
    }

    public ZClassInfoDecorator getDecorator() {
        return decorator;
    }

    public void setDecorator(ZClassInfoDecorator decorator) {
        this.decorator = decorator;
    }

    public ZIClassInfo getClassInfo() {
        return classInfo;
    }
}
