/*
 * Copyright 2008 Gerd Ziegler (www.gerdziegler.de) Licensed under the Apache
 * License, Version 2.0 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.apache.org/licenses/LICENSE-2.0 Unless required by applicable law
 * or agreed to in writing, software distributed under the License is
 * distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the specific language
 * governing permissions and limitations under the License. @author
 * www.gerdziegler.de
 */
package org.ztest.output.classinfo;

import java.awt.Color;
import java.io.Writer;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.apache.log4j.Logger;
import org.ztest.analyze.ZAnalyzeClassSetDependency;
import org.ztest.classifier.ZClassifier;
import org.ztest.classifier.ZClassifyByLocation;
import org.ztest.classinfo.ZIClassFilter;
import org.ztest.classinfo.ZIClassInfo;
import org.ztest.classinfo.ZIClassSetEdgeMeta;
import org.ztest.filter.ZClassFilterAcceptAll;
import org.ztest.graph.ZIEdge;
import org.ztest.graph.ZIGraph;
import org.ztest.jdom.ZWriteXML;
import org.ztest.output.graphml.ZGraphMLWriter;
import org.ztest.output.graphml.ZIGraphMLDecorator;
import org.ztest.output.graphml.ZIGraphMLEdge;
import org.ztest.output.graphml.ZIGraphMLGroupNode;
import org.ztest.output.graphml.ZIGraphMLNode;
import org.ztest.output.graphml.style.ZEdgeStyle;
import org.ztest.output.graphml.style.ZGroupNodeStyle;
import org.ztest.output.graphml.style.ZNodeStyle;
import org.ztest.output.graphml.style.ZStyleUtil;
import org.ztest.sets.ZIClassSet;

public class ZLocationInfoWriter {
	static final Logger log = Logger.getLogger(ZLocationInfoWriter.class);

	private final ZIClassInfo classInfo;

	private ZIClassFilter classFilter = new ZClassFilterAcceptAll();

	private boolean cyclesOnly = false;

	public ZLocationInfoWriter(final ZIClassInfo classInfo,
			ZIClassFilter classFilter) {
		super();
		this.classInfo = classInfo;
		this.classFilter = classFilter;
	}

	public ZLocationInfoWriter(final ZIClassInfo classInfo) {
		super();
		this.classInfo = classInfo;
	}

	public void write(Writer writer) throws Exception {
		log.debug("computing location dependencies...");
		Set<ZIClassSet> classification = ZClassifier.classify(classInfo
				.getClasses(classFilter), new ZClassifyByLocation(classInfo));

		log.debug("exporting location dependencies...");
		ZIGraph<ZIClassSet, ZIClassSetEdgeMeta, ZWriteXML> graph = ZAnalyzeClassSetDependency
				.computeDependencyGraph(classInfo, classification);

		exportGraphML(graph, writer);
	}

	static int isInCycle(List<Set<ZIClassSet>> cycles, String name) {
		for (int i = 0; i < cycles.size(); i++) {
			Set<ZIClassSet> cycle = cycles.get(i);
			for (ZIClassSet cs : cycle) {
				if (cs.getName().equals(name)) {
					return i;
				}
			}
		}
		return -1;
	}

	static int isInCycle(List<Set<ZIClassSet>> cycles, String src, String tgt) {
		int i1 = isInCycle(cycles, src);
		int i2 = isInCycle(cycles, tgt);
		return i1 == i2 ? i1 : -1;
	}

	private void exportGraphML(
			ZIGraph<ZIClassSet, ZIClassSetEdgeMeta, ZWriteXML> graph,
			Writer writer) throws Exception {

		final List<Color> cycleColors = new ArrayList<Color>();
		final List<Set<ZIClassSet>> cycles = graph.getCycles();
		for (int i = 0; i < cycles.size(); i++) {
			cycleColors.add(ZStyleUtil.getRandomColor(190, 225, 190, 225, 170,
					190));
		}

		ZIGraphMLDecorator decorator = new ZIGraphMLDecorator() {
			public ZNodeStyle getStyle(ZIGraphMLNode node, ZNodeStyle style)
					throws Exception {
				int cycle = ZLocationInfoWriter.isInCycle(cycles, node.getId());
				if (cycle >= 0) {
					Color col = cycleColors.get(cycle);
					style.getFillStyle().setColors(col.getRed(),
							col.getGreen(), col.getBlue());
				}
				style.getNodeLabelStyle().setText(node.getId());
				return style;
			}

			public ZGroupNodeStyle getStyle(ZIGraphMLGroupNode group,
					ZGroupNodeStyle style) throws Exception {
				style.getNodeLabelStyle().setText(group.getId());
				return style;
			}

			public ZEdgeStyle getStyle(ZIGraphMLEdge edge, ZEdgeStyle style)
					throws Exception {

				int cycle = ZLocationInfoWriter.isInCycle(cycles, edge
						.getSourceId(), edge.getTargetId());
				if (cycle >= 0) {
					style.getLineStyle().setWidth(3);
					// style.getLineStyle().setType(ZLineStyle.LINE_TYPE_DASHED);
					Color col = cycleColors.get(cycle);
					style.getLineStyle().setColor(col.getRed(), col.getGreen(),
							col.getBlue());
				}

				style
						.setText(edge.getSourceId() + " --> "
								+ edge.getTargetId());
				return style;
			}
		};

		Set<ZIGraphMLNode> nodes = new HashSet<ZIGraphMLNode>();
		for (ZIClassSet cs : graph.getNodeSet()) {
			if (!cyclesOnly || isInCycle(cycles, cs.getName()) >= 0) {
				nodes.add(createGraphMLNode(cs));
			}
		}

		Set<ZIGraphMLEdge> edges = new HashSet<ZIGraphMLEdge>();
		for (ZIEdge<ZIClassSet, ZIClassSetEdgeMeta> edge : graph.getEdgeSet()) {
			if (!cyclesOnly
					|| isInCycle(cycles, edge.getSource().getName(), edge
							.getTarget().getName()) >= 0) {
				edges.add(createGraphMLEdge(edge));
			}
		}

		ZGraphMLWriter exporter = new ZGraphMLWriter(nodes, edges, decorator);
		exporter.write(writer);
	}

	private static ZIGraphMLEdge createGraphMLEdge(
			final ZIEdge<ZIClassSet, ZIClassSetEdgeMeta> edge) {
		ZIGraphMLEdge ret = new ZIGraphMLEdge() {
			public String getSourceId() throws Exception {
				return edge.getSource().getName();
			}

			public String getTargetId() throws Exception {
				return edge.getTarget().getName();
			}
		};
		return ret;
	}

	private static ZIGraphMLNode createGraphMLNode(final ZIClassSet classSet) {
		ZIGraphMLNode ret = new ZIGraphMLNode() {
			public String getId() {
				return classSet.getName();
			}
		};
		return ret;
	}

	public ZIClassFilter getClassFilter() {
		return classFilter;
	}

	public void setClassFilter(ZIClassFilter classFilter) {
		this.classFilter = classFilter;
	}

	public ZIClassInfo getClassInfo() {
		return classInfo;
	}

	public boolean isCyclesOnly() {
		return cyclesOnly;
	}

	public void setCyclesOnly(boolean cyclesOnly) {
		this.cyclesOnly = cyclesOnly;
	}

}
