/*
 * Copyright 2008 Gerd Ziegler (www.gerdziegler.de) Licensed under the Apache
 * License, Version 2.0 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.apache.org/licenses/LICENSE-2.0 Unless required by applicable law
 * or agreed to in writing, software distributed under the License is
 * distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the specific language
 * governing permissions and limitations under the License. @author
 * www.gerdziegler.de
 */
package org.ztest.output.graphml;

import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.geom.Rectangle2D;
import java.awt.image.BufferedImage;
import java.io.Writer;
import java.util.Set;

import org.jdom.Document;
import org.jdom.Element;
import org.jdom.Namespace;
import org.jdom.Text;
import org.jdom.output.Format;
import org.jdom.output.XMLOutputter;
import org.ztest.output.graphml.style.ZBorderInsetsStyle;
import org.ztest.output.graphml.style.ZBorderStyle;
import org.ztest.output.graphml.style.ZEdgeStyle;
import org.ztest.output.graphml.style.ZFillStyle;
import org.ztest.output.graphml.style.ZGroupNodeStyle;
import org.ztest.output.graphml.style.ZInsetsStyle;
import org.ztest.output.graphml.style.ZNodeLabelStyle;
import org.ztest.output.graphml.style.ZNodeStyle;
import org.ztest.output.graphml.style.ZShapeStyle;
import org.ztest.output.graphml.style.ZStateStyle;

public class ZGraphMLWriter {
    private final ZIGraphMLConnector connector;

    private final Graphics graphics;

    private final Namespace namespace = Namespace.getNamespace("http://graphml.graphdrawing.org/xmlns/graphml");

    private final Namespace yns = Namespace.getNamespace("y", "http://www.yworks.com/xml/graphml");

    private Set<ZIGraphMLNode> nodes;

    private Set<ZIGraphMLEdge> edges;

    private ZIGraphMLDecorator decorator;

    // private final Namespace xsi = Namespace.getNamespace("xsi",
    // "http://www.w3.org/2001/XMLSchema-instance");

    public ZGraphMLWriter(ZIGraphMLConnector connector) {
        this.connector = connector;
        BufferedImage img = new BufferedImage(400, 40, BufferedImage.TYPE_INT_RGB);
        graphics = img.getGraphics();
    }

    public ZGraphMLWriter(final Set<ZIGraphMLNode> nodes, final Set<ZIGraphMLEdge> edges, final ZIGraphMLDecorator decorator) {
        this(new ZIGraphMLConnector() {

            public ZIGraphMLDecorator getGraphMLDecorator()
                throws Exception {
                return decorator;
            }

            public Set<ZIGraphMLEdge> getGraphMLEdges()
                throws Exception {
                return edges;
            }

            public Set<ZIGraphMLNode> getGraphMLNodes()
                throws Exception {
                return nodes;
            }
        });
    }

    public void write(Writer w)
        throws Exception {

        nodes = connector.getGraphMLNodes();
        edges = connector.getGraphMLEdges();
        decorator = connector.getGraphMLDecorator();
        try {

            Element root = new Element("graphml", namespace);
            // root.addNamespaceDeclaration(xsi);
            root.addNamespaceDeclaration(yns);
            // root.setAttribute("schemaLocation",
            // "http://graphml.graphdrawing.org/xmlns/graphml
            // http://www.yworks.com/xml/schema/graphml/1.0/ygraphml.xsd", xsi);

            root.addContent(new Element("key", namespace).setAttribute("for", "node").setAttribute("id", "d0").setAttribute("yfiles.type", "nodegraphics"));
            root.addContent(new Element("key", namespace).setAttribute("for", "node").setAttribute("id", "d1").setAttribute("attr.name", "description").setAttribute("attr.type", "string"));
            root.addContent(new Element("key", namespace).setAttribute("for", "edge").setAttribute("id", "d2").setAttribute("yfiles.type", "edgegraphics"));
            root.addContent(new Element("key", namespace).setAttribute("for", "edge").setAttribute("id", "d3").setAttribute("attr.name", "description").setAttribute("attr.type", "string"));
            root.addContent(new Element("key", namespace).setAttribute("for", "graphml").setAttribute("id", "d4").setAttribute("yfiles.type", "resources"));

            root.addContent(toElement());

            Document doc = new Document(root);
            XMLOutputter outputter = new XMLOutputter(Format.getPrettyFormat().setExpandEmptyElements(false));
            outputter.output(doc, w);
            w.flush();
        } finally {
            nodes = null;
            edges = null;
            decorator = null;
        }
    }

    private Element toElement()
        throws Exception {

        Element ret = new Element("graph", namespace);
        ret.setAttribute("edgedefault", "directed");
        ret.setAttribute("id", "G");
        ret.setAttribute("parse.edges", String.valueOf(edges.size()));
        ret.setAttribute("parse.nodes", String.valueOf(nodes.size()));

        for (ZIGraphMLNode node : nodes) {
            ret.addContent(toElement(node));
        }

        for (ZIGraphMLEdge edge : edges) {
            ret.addContent(toElement(edge));
        }

        return ret;
    }

    private Element toElement(ZIGraphMLNode node)
        throws Exception {
        if (node instanceof ZIGraphMLGroupNode) {
            return toFolderElement((ZIGraphMLGroupNode)node);
        } else {
            return toNodeElement(node);
        }
    }

    private Element toNodeElement(ZIGraphMLNode node)
        throws Exception {

        Element ret = new Element("node", namespace);
        ret.setAttribute("id", node.getId());
        Element d0 = new Element("data", namespace).setAttribute("key", "d0");
        ret.addContent(d0);

        ZNodeStyle nodeStyle = decorator.getStyle(node, new ZNodeStyle());

        Element shapeNode = new Element("ShapeNode", yns);
        d0.addContent(shapeNode);

        shapeNode.addContent(toShapeElement(nodeStyle.getShapeStyle()));
        shapeNode.addContent(toElement(nodeStyle.getFillStyle()));
        shapeNode.addContent(toNodeLabelElement(nodeStyle.getNodeLabelStyle()));
        shapeNode.addContent(toGeometryElement(nodeStyle.getNodeLabelStyle()));
        shapeNode.addContent(toElement(nodeStyle.getBorderStyle()));
        shapeNode.addContent(toNodeBoundsElement());
        return ret;
    }

    private Element toNodeBoundsElement() {
        Element nodeBoundsElement = new Element("NodeBounds");
        nodeBoundsElement.setAttribute("considerNodeLabelSize", "true");
        return nodeBoundsElement;
    }

    private Element toFolderElement(ZIGraphMLGroupNode node)
        throws Exception {

        Element ret = new Element("node", namespace);
        ret.setAttribute("id", node.getId());
        ret.setAttribute("yfiles.foldertype", "group");

        Element d0 = new Element("data", namespace).setAttribute("key", "d0");
        ret.addContent(d0);

        ZGroupNodeStyle folderStyle = decorator.getStyle(node, new ZGroupNodeStyle());

        Element autoBoundsNode = new Element("ProxyAutoBoundsNode", yns);
        d0.addContent(autoBoundsNode);

        Element realizersNode = new Element("Realizers", yns);
        autoBoundsNode.addContent(realizersNode);
        realizersNode.setAttribute("active", "0");
        folderStyle.getInsetsStyle().setBottom(15);
        folderStyle.getInsetsStyle().setLeft(50);
        folderStyle.getInsetsStyle().setRight(15);
        folderStyle.getInsetsStyle().setTop(15);
        folderStyle.getStateStyle().setClosed(true);
        realizersNode.addContent(toGroupNode(folderStyle));

        folderStyle.getStateStyle().setClosed(false);
        realizersNode.addContent(toGroupNode(folderStyle));

        Element graphNode = new Element("graph", namespace);
        graphNode.setAttribute("edgedefault", "directed");
        graphNode.setAttribute("id", node.getId() + ":");
        ret.addContent(graphNode);
        for (ZIGraphMLNode n : node.getNodes()) {
            graphNode.addContent(toElement(n));
        }

        return ret;
    }

    private Element toGroupNode(ZGroupNodeStyle folderStyle) {
        Element groupNode = new Element("GroupNode", yns);

        groupNode.addContent(toShapeElement(folderStyle.getShapeStyle()));
        groupNode.addContent(toElement(folderStyle.getFillStyle()));
        groupNode.addContent(toNodeLabelElement(folderStyle.getNodeLabelStyle()));
        groupNode.addContent(toGeometryElement(folderStyle.getNodeLabelStyle()));
        groupNode.addContent(toElement(folderStyle.getBorderStyle()));
        groupNode.addContent(toInsetsElement(folderStyle.getInsetsStyle()));
        groupNode.addContent(toBorderInsetsElement(folderStyle.getBorderInsetsStyle()));
        groupNode.addContent(toStateElement(folderStyle.getStateStyle()));
        groupNode.addContent(toNodeBoundsElement());

        return groupNode;
    }

    private Element toInsetsElement(ZInsetsStyle insetsStyle) {
        Element ret = new Element("Insets", yns);
        ret.setAttribute("bottom", String.valueOf(insetsStyle.getBottom()));
        ret.setAttribute("left", String.valueOf(insetsStyle.getLeft()));
        ret.setAttribute("right", String.valueOf(insetsStyle.getRight()));
        ret.setAttribute("top", String.valueOf(insetsStyle.getTop()));
        return ret;
    }

    private Element toBorderInsetsElement(ZBorderInsetsStyle borderStyle) {
        Element ret = new Element("BorderInsets", yns);
        ret.setAttribute("bottom", String.valueOf(borderStyle.getBottom()));
        ret.setAttribute("left", String.valueOf(borderStyle.getLeft()));
        ret.setAttribute("right", String.valueOf(borderStyle.getRight()));
        ret.setAttribute("top", String.valueOf(borderStyle.getTop()));
        return ret;
    }

    private Element toStateElement(ZStateStyle stateStyle) {
        Element ret = new Element("State", yns);
        ret.setAttribute("closed", String.valueOf(stateStyle.isClosed()));
        ret.setAttribute("innerGraphDisplayEnabled", String.valueOf(stateStyle.isInnerGraphDisplayEnabled()));
        return ret;
    }

    private Element toElement(ZBorderStyle borderStyle) {
        Element ret = new Element("BorderStyle", yns);
        ret.setAttribute("color", borderStyle.getLineStyle().getColor());
        ret.setAttribute("type", borderStyle.getLineStyle().getType());
        ret.setAttribute("width", String.valueOf(borderStyle.getLineStyle().getWidth()));
        return ret;
    }

    private Element toElement(ZFillStyle fillStyle) {
        Element ret = new Element("Fill", yns);
        ret.setAttribute("color", fillStyle.getColor());
        ret.setAttribute("color2", fillStyle.getColor2());
        ret.setAttribute("transparent", String.valueOf(fillStyle.isTransparent()));
        return ret;
    }

    private Element toNodeLabelElement(ZNodeLabelStyle textStyle) {
        String txt = textStyle.getText();
        Font font = textStyle.getFont();

        Element ret = new Element("NodeLabel", yns);
        ret.addContent(new Text(txt));
        ret.setAttribute("alignment", "center");
        ret.setAttribute("autoSizePolicy", "content");
        ret.setAttribute("fontFamily", font.getName());
        ret.setAttribute("fontSize", String.valueOf(font.getSize()));
        if (font.isPlain()) {
            ret.setAttribute("fontStyle", "plain");
        } else if (font.isBold()) {
            ret.setAttribute("fontStyle", "bold");
        } else if (font.isItalic()) {
            ret.setAttribute("fontStyle", "italic");
        }
        ret.setAttribute("modelName", "internal");
        ret.setAttribute("modelPosition", "c");
        return ret;
    }

    private Element toGeometryElement(ZNodeLabelStyle textStyle) {
        String txt = textStyle.getText();
        Font font = textStyle.getFont();
        FontMetrics fm = graphics.getFontMetrics(font);
        Rectangle2D rect = fm.getStringBounds(txt, graphics);

        Element ret = new Element("Geometry", yns);
        ret.setAttribute("height", String.valueOf(rect.getHeight() + 10));
        ret.setAttribute("width", String.valueOf(rect.getWidth() + 20));
        return ret;
    }

    private Element toShapeElement(ZShapeStyle shapeStyle) {
        Element ret = new Element("Shape", yns);
        ret.setAttribute("type", shapeStyle.getType());
        return ret;
    }

    private Element toElement(ZIGraphMLEdge edge)
        throws Exception {

        ZEdgeStyle edgeStyle = decorator.getStyle(edge, new ZEdgeStyle());

        Element ret = new Element("edge", namespace);
        ret.setAttribute("id", edge.getSourceId() + "-" + edge.getTargetId());
        ret.setAttribute("source", edge.getSourceId());
        ret.setAttribute("target", edge.getTargetId());
        Element d2 = new Element("data", namespace);
        d2.setAttribute("key", "d2");
        ret.addContent(d2);

        Element d3 = new Element("data", namespace);
        ret.addContent(d3);
        d3.setAttribute("key", "d3");
        d3.addContent(new Text(edgeStyle.getText()));

        Element polyEdge = new Element("PolyLineEdge", yns);
        d2.addContent(polyEdge);

        Element path = new Element("Path", yns);
        polyEdge.addContent(path);

        Element lineStyle = new Element("LineStyle", yns);
        polyEdge.addContent(lineStyle);
        lineStyle.setAttribute("color", edgeStyle.getLineStyle().getColor());
        lineStyle.setAttribute("type", edgeStyle.getLineStyle().getType());
        lineStyle.setAttribute("width", String.valueOf(edgeStyle.getLineStyle().getWidth()));

        Element arrowsEdge = new Element("Arrows", yns);
        polyEdge.addContent(arrowsEdge);
        arrowsEdge.setAttribute("source", edgeStyle.getArrow1());
        arrowsEdge.setAttribute("target", edgeStyle.getArrow2());

        Element bendStyle = new Element("BendStyle", yns);
        polyEdge.addContent(bendStyle);
        bendStyle.setAttribute("smoothed", "false");

        return ret;
    }

}
