/*
 * Copyright 2008 Gerd Ziegler (www.gerdziegler.de) Licensed under the Apache
 * License, Version 2.0 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.apache.org/licenses/LICENSE-2.0 Unless required by applicable law
 * or agreed to in writing, software distributed under the License is
 * distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the specific language
 * governing permissions and limitations under the License. @author
 * www.gerdziegler.de
 */
package org.ztest.tests;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.log4j.Logger;
import org.jdom.Element;
import org.jdom.Text;
import org.ztest.classinfo.ZIClassEdge;
import org.ztest.classinfo.ZIClassInfo;
import org.ztest.classinfo.ZIClassSetEdgeMeta;
import org.ztest.classinfo.ZIPath;
import org.ztest.classinfo.impl.ZDependencyEdgeMeta;
import org.ztest.graph.ZIEdge;
import org.ztest.graph.ZIGraph;
import org.ztest.graph.jgrapht.ZJGraphTGraph;
import org.ztest.jdom.ZWriteXML;
import org.ztest.jdom.ZXMLUtil;
import org.ztest.sets.ZIClassSet;

/**
 * Specify dependencies between class-sets. Any unspecified dependency is
 * considered invalid and fails the test. Checks all paths from one member of a
 * class-set to the other. If path goes through classes that are not contained
 * in any class-set, all class-sets transitively reachable are computed. Test
 * fails if unallowed class-sets are reachable.
 *
 * @author www.gerdziegler.de
 */
public class ZDependencyTest implements ZITest
{
  static Logger log = Logger.getLogger(ZDependencyTest.class);

  private ZIClassInfo classInfo;

  private final Set<String> unclassified = new HashSet<String>();

  private FailMode failMode = FailMode.LOOSE;

  private boolean xmlShowClassSets = false;

  private boolean xmlShowUnclassified = false;

  private final ZIGraph<ZIClassSet, ZIClassSetEdgeMeta, ZWriteXML> allowedDependencyGraph = new ZJGraphTGraph<ZIClassSet, ZIClassSetEdgeMeta, ZWriteXML>();

  private final ZIGraph<ZIClassSet, ZIClassSetEdgeMeta, ZWriteXML> allowedDependencyGraphWithIncludes = new ZJGraphTGraph<ZIClassSet, ZIClassSetEdgeMeta, ZWriteXML>();

  private final ZIGraph<ZIClassSet, ZIClassSetEdgeMeta, ZWriteXML> actualDependencyGraphStrict = new ZJGraphTGraph<ZIClassSet, ZIClassSetEdgeMeta, ZWriteXML>();

  private final ZIGraph<ZIClassSet, ZIClassSetEdgeMeta, ZWriteXML> actualDependencyGraphLoose = new ZJGraphTGraph<ZIClassSet, ZIClassSetEdgeMeta, ZWriteXML>();

  private final List<ZDependencyFailure> failuresLoose = new ArrayList<ZDependencyFailure>();

  private final List<ZDependencyFailure> failuresStrict = new ArrayList<ZDependencyFailure>();

  private Map<String, Set<ZIClassSet>> class2ClassSetMap;

  public enum FailMode {
    STRICT("strict"), LOOSE("loose");

    private final String text;


    private FailMode(final String text)
    {
      this.text = text;
    }


    public String getText()
    {
      return text;
    }
  }


  // /**
  // * calls add with acceptCollisions true
  // *
  // * @param t
  // * @throws Exception
  // */
  // public void addAll(ZDependencyTest t)
  // throws Exception {
  // addAll(t, true);
  // }

  /**
   * adds an test to this one by adding all classsets, edges and metadata.
   *
   * @param t
   * @param acceptCollisions
   * @throws Exception
   */
  // public void addAll(ZDependencyTest t, boolean acceptCollisions)
  // throws Exception {
  // for (ZIClassSet cs : t.getAllowedDependencyGraph().getNodeSet()) {
  // if (allowedDependencyGraph.isNode(cs)) {
  // if (!acceptCollisions) {
  // throw new Exception("class-set already exists: " + cs);
  // }
  // } else {
  // allowedDependencyGraph.addNode(cs);
  // actualDependencyGraphStrict.addNode(cs);
  // actualDependencyGraphLoose.addNode(cs);
  // }
  // }
  // for (ZIEdge<ZIClassSet, ZIClassSetEdgeMeta> edge :
  // t.getAllowedDependencyGraph().getEdgeSet()) {
  // ZIEdge<ZIClassSet, ZIClassSetEdgeMeta> newEdge =
  // allowedDependencyGraph.getEdge(edge.getSource(), edge.getTarget());
  // if (newEdge == null) {
  // newEdge = allowedDependencyGraph.addEdge(edge.getSource(),
  // edge.getTarget());
  // }
  // newEdge.setMetaData(edge.getMetaData());
  // }
  // }
  public void init(ZIClassInfo classInfo) throws Exception
  {
    this.classInfo = classInfo;
  }


  public void run() throws Exception
  {
    createAllowedDependencyGraphWithIncludes();

    check();

    if (xmlShowUnclassified)
    {
      Set<String> scanned = classInfo.getScannedClasses();
      unclassified.addAll(scanned);
      for (ZIClassSet set : allowedDependencyGraph.getNodeSet())
      {
        unclassified.removeAll(set.getClasses());
      }
    }
  }


  /**
   * nested set inherits from including set
   *
   * @throws Exception
   */
  private void createAllowedDependencyGraphWithIncludes() throws Exception
  {
    for (ZIClassSet set1 : allowedDependencyGraph.getNodeSet())
    {
      allowedDependencyGraphWithIncludes.addNode(set1);
    }

    for (ZIClassSet set1 : allowedDependencyGraph.getNodeSet())
    {
      for (ZIClassSet set2 : allowedDependencyGraph.getNodeSet())
      {
        if (set1 == set2)
        {
          continue;
        }

        if (allowedDependencyGraph.isEdge(set1, set2))
        {
          allowedDependencyGraphWithIncludes.addEdge(set1, set2);
        }
        else if (set1.getClasses().containsAll(set2.getClasses()))
        {
          allowedDependencyGraphWithIncludes.addEdge(set1, set2);
          allowedDependencyGraphWithIncludes.addEdge(set2, set1);
          for (ZIClassSet p : allowedDependencyGraph.getPredecessors(set1, null, false))
          {
            allowedDependencyGraphWithIncludes.addEdge(p, set2);
          }
          for (ZIClassSet s : allowedDependencyGraph.getSuccessors(set1, null, false))
          {
            allowedDependencyGraphWithIncludes.addEdge(set2, s);
          }
        }
      }
    }
  }


  public void addDependency(ZIClassSet set1, ZIClassSet... sets) throws Exception
  {
    for (ZIClassSet set2 : sets)
    {
      if (!allowedDependencyGraph.isNode(set1))
      {
        throw new Exception("explicitly add set to test before declaring dependency: " + set1.getName());
      }
      if (!allowedDependencyGraph.isNode(set2))
      {
        throw new Exception("explicitly add set to test before declaring dependency: " + set2.getName());
      }
      allowedDependencyGraph.addEdge(set1, set2);
    }
  }


  public String toXML() throws Exception
  {
    return ZXMLUtil.toXML(toElement());
  }


  public Element toElement() throws Exception
  {
    Element ret = new Element("dependency-test");
    ret.addContent(ZTestUtil.toElement(getFailures()));

    if (xmlShowClassSets)
    {
      Element layersNode = new Element("layers");
      ret.addContent(layersNode);
      for (ZIClassSet layer : allowedDependencyGraph.getNodeSet())
      {
        layersNode.addContent(layer.toElement());
      }
    }

    if (xmlShowUnclassified)
    {
      Element unclassifiedNode = new Element("unclassified");
      ret.addContent(unclassifiedNode);
      for (String t : unclassified)
      {
        Element classNode = new Element("class");
        classNode.addContent(new Text(t.toString()));
        unclassifiedNode.addContent(classNode);
      }
    }

    return ret;
  }


  public Set<String> getUnclassified()
  {
    return unclassified;
  }


  // private Element toElementLayer(String name, Set<String> classes)
  // {
  // Element root = new Element("class-set");
  // root.setAttribute("name", name);
  // root.setAttribute("count", String.valueOf(classes.size()));
  //
  // List<String> sorted = ZTestUtil.sort(classes);
  // for (String s : sorted)
  // {
  // Element classNode = new Element("class");
  // classNode.addContent(new Text(s.toString()));
  // root.addContent(classNode);
  // }
  // return root;
  // }

  public void addClassSet(ZIClassSet set) throws Exception
  {
    if (allowedDependencyGraph.isNode(set))
    {
      throw new Exception("class-set already registered: " + set.getName());
    }
    allowedDependencyGraph.addNode(set);
    actualDependencyGraphStrict.addNode(set);
    actualDependencyGraphLoose.addNode(set);
  }


  private void check() throws Exception
  {
    class2ClassSetMap = computeClass2ClassSetMap(allowedDependencyGraphWithIncludes.getNodeSet());
    Map<String, Set<String>> successorCache = new HashMap<String, Set<String>>();

    for (ZIClassEdge e : classInfo.getEdges())
    {
      String source = e.getSource();
      String classOnPath = e.getTarget();

      if (source.equals(classOnPath))
      {
        continue;
      }

      Set<ZIClassSet> l1 = class2ClassSetMap.get(source);
      if (l1 == null)
      {
        continue;
      }

      Set<String> successors = getLayeredSuccessors(class2ClassSetMap, successorCache, classOnPath);

      for (String target : successors)
      {
        checkAllowed(class2ClassSetMap, source, classOnPath, target);
      }
    }
  }


  private Set<String> getLayeredSuccessors(Map<String, Set<ZIClassSet>> layerMap, Map<String, Set<String>> layeredSuccessorCache, String name) throws Exception
  {
    Set<String> ret = layeredSuccessorCache.get(name);
    if (ret != null)
    {
      return ret;
    }

    if (layerMap.containsKey(name))
    {
      ret = new HashSet<String>();
      ret.add(name);
      layeredSuccessorCache.put(name, ret);
      return ret;
    }

    ret = new HashSet<String>();
    layeredSuccessorCache.put(name, ret);
    Set<String> successors = classInfo.getSuccessors(name, false);
    for (String crt : successors)
    {
      if (!crt.equals(name))
      {
        Set<String> crtSet = getLayeredSuccessors(layerMap, layeredSuccessorCache, crt);
        ret.addAll(crtSet);
      }
    }

    ret.remove(name);

    return ret;
  }


  private void checkAllowed(Map<String, Set<ZIClassSet>> class2ClassSetMap, String source, String classOnPath, String target) throws Exception
  {
    Set<ZIClassSet> classSets1 = class2ClassSetMap.get(source);
    Set<ZIClassSet> classSets2 = class2ClassSetMap.get(target);

    boolean allowStrict = true;
    boolean allowLoose = false;
    if (!classSets1.isEmpty() && !classSets2.isEmpty() && !classSets1.equals(classSets2))
    {
      for (ZIClassSet set1 : classSets1)
      {
        for (ZIClassSet set2 : classSets2)
        {
          if (set1.getName().equals(set2.getName()))
          {
            allowLoose = true;
          }
          else if (allowedDependencyGraphWithIncludes.isEdge(set1, set2))
          {
            allowLoose = true;
          }
          else
          {
            allowStrict = false;
          }
        }
      }
    }
    else
    {
      allowLoose = true;
    }

    ZIPath path = null;
    if (!allowStrict || !allowLoose)
    {
      path = classInfo.getShortestPath(classOnPath, target);
      ZIClassEdge edge = classInfo.getEdge(source, classOnPath);
      path.getEdges().add(0, edge);
      ZDependencyFailure failure = new ZDependencyFailure(path, class2ClassSetMap);

      if (!allowStrict)
      {
        failuresStrict.add(failure);
      }
      if (!allowLoose)
      {
        failuresLoose.add(failure);
      }
    }

    for (ZIClassSet set1 : classSets1)
    {
      for (ZIClassSet set2 : classSets2)
      {
        if (!set1.getName().equals(set2.getName()))
        {
          if (allowedDependencyGraphWithIncludes.isEdge(set1, set2))
          {
            actualDependencyGraphLoose.addEdge(set1, set2);
            actualDependencyGraphStrict.addEdge(set1, set2);
          }
          else
          {
            if (!allowLoose)
            {
              ZIEdge<ZIClassSet, ZIClassSetEdgeMeta> edge = actualDependencyGraphLoose.addEdge(set1, set2);
              if (edge.getMetaData() == null)
              {
                edge.setMetaData(new ZDependencyEdgeMeta());
              }
              edge.getMetaData().getPaths().add(path);
            }
            if (!allowStrict)
            {
              ZIEdge<ZIClassSet, ZIClassSetEdgeMeta> edge = actualDependencyGraphStrict.addEdge(set1, set2);
              if (edge.getMetaData() == null)
              {
                edge.setMetaData(new ZDependencyEdgeMeta());
              }
              edge.getMetaData().getPaths().add(path);
            }
          }
        }
      }
    }
  }


  // private AllowMode isAllowed(Map<String, Set<ZIClassSet>> class2ClassSetMap,
  // String source, String target) throws Exception
  // {
  // Set<ZIClassSet> classSets1 = class2ClassSetMap.get(source);
  // Set<ZIClassSet> classSets2 = class2ClassSetMap.get(target);
  // if (classSets1.isEmpty() || classSets2.isEmpty() ||
  // classSets1.equals(classSets2))
  // {
  // return AllowMode.STRICT;
  // }
  // boolean allowStrict = true;
  // boolean allowLoose = false;
  // for (ZIClassSet set1 : classSets1)
  // {
  // for (ZIClassSet set2 : classSets2)
  // {
  // if (set1.equals(set2) || allowedDependencyGraph.isEdge(set1, set2))
  // {
  // allowLoose = true;
  // }
  // else
  // {
  // allowStrict = false;
  // }
  // }
  // }
  //
  // if (allowStrict)
  // {
  // return AllowMode.STRICT;
  // }
  // else if (allowLoose)
  // {
  // return AllowMode.LOOSE;
  // }
  // else
  // {
  // return AllowMode.FORBIDDEN;
  // }
  // }

  private Map<String, Set<ZIClassSet>> computeClass2ClassSetMap(Set<ZIClassSet> classSets) throws Exception
  {
    Map<String, Set<ZIClassSet>> ret = new HashMap<String, Set<ZIClassSet>>();
    for (ZIClassSet set : classSets)
    {
      for (String clazz : set.getClasses())
      {
        Set<ZIClassSet> dependencySet = ret.get(clazz);
        if (dependencySet == null)
        {
          dependencySet = new HashSet<ZIClassSet>();
          ret.put(clazz, dependencySet);
        }
        dependencySet.add(set);
      }
    }
    return ret;
  }


  public ZIClassInfo getClassInfo()
  {
    return classInfo;
  }


  public List<ZDependencyFailure> getFailures()
  {
    if (failMode == FailMode.LOOSE)
    {
      return failuresLoose;
    }
    else if (failMode == FailMode.STRICT)
    {
      return failuresStrict;
    }
    throw new RuntimeException("unsupported fail mode " + failMode);
  }


  public boolean isXmlShowClassSets()
  {
    return xmlShowClassSets;
  }


  public void setXmlShowClassSets(boolean xmlShowLayers)
  {
    this.xmlShowClassSets = xmlShowLayers;
  }


  public boolean isXmlShowUnclassified()
  {
    return xmlShowUnclassified;
  }


  public void setXmlShowUnclassified(boolean xmlShowUnclassified)
  {
    this.xmlShowUnclassified = xmlShowUnclassified;
  }


  public ZIGraph<ZIClassSet, ZIClassSetEdgeMeta, ZWriteXML> getAllowedDependencyGraph()
  {
    return allowedDependencyGraph;
  }


  public Set<ZIClassSet> getClassSets() throws Exception
  {
    return allowedDependencyGraph.getNodeSet();
  }


  /**
   * run test first to get this filled with data
   *
   * @return
   */
  public ZIGraph<ZIClassSet, ZIClassSetEdgeMeta, ZWriteXML> getActualDependencyGraphStrict()
  {
    return actualDependencyGraphStrict;
  }


  public ZIGraph<ZIClassSet, ZIClassSetEdgeMeta, ZWriteXML> getActualDependencyGraphLoose()
  {
    return actualDependencyGraphLoose;
  }


  public Map<String, Set<ZIClassSet>> getClass2ClassSetMap()
  {
    return class2ClassSetMap;
  }


  public List<ZDependencyFailure> getFailuresLoose()
  {
    return failuresLoose;
  }


  public List<ZDependencyFailure> getFailuresStrict()
  {
    return failuresStrict;
  }


  public FailMode getFailMode()
  {
    return failMode;
  }


  public void setFailMode(FailMode failMode)
  {
    this.failMode = failMode;
  }


  public ZIGraph<ZIClassSet, ZIClassSetEdgeMeta, ZWriteXML> getAllowedDependencyGraphWithIncludes()
  {
    return allowedDependencyGraphWithIncludes;
  }
}
